"""
survey_publisher_v1
=========================

This module loads the survey data from before 2022 from an excel file.
Missing info is filled in from the survey db for some questions.
Registered as click cli command when installing compendium-v2.

"""
from __future__ import annotations
import logging
import math
import click

from sqlalchemy import select

import compendium_v2
from compendium_v2.environment import setup_logging
from compendium_v2.background_task import parse_excel_data
from compendium_v2.config import load
from compendium_v2.db import db, model
from compendium_v2.survey_db import model as survey_model
from compendium_v2.publishers import helpers

setup_logging()

logger = logging.getLogger('survey-publisher-v1')


def db_budget_migration(nren_dict):
    # move data from Survey DB budget table
    data = db.session.scalars(select(survey_model.Nrens))
    for nren in data:
        for budget in nren.budgets:
            abbrev = nren.abbreviation.upper()
            year = budget.year

            if (year == 2013 and abbrev == 'JISC') or (year == 2014 and abbrev == 'GARR'):
                # these are obviously invalid datapoints
                continue

            if float(budget.budget) > 200:
                logger.warning(f'Incorrect Data: {abbrev} has budget set >200M EUR for {year}. ({budget.budget})')

            if abbrev not in nren_dict:
                logger.warning(f'{abbrev} unknown. Skipping.')
                continue

            budget_entry = model.BudgetEntry(
                nren=nren_dict[abbrev],
                nren_id=nren_dict[abbrev].id,
                budget=float(budget.budget),
                year=year
            )
            db.session.merge(budget_entry)

    # Import the data from excel sheet to database
    exceldata = parse_excel_data.fetch_budget_excel_data()

    for abbrev, budget, year in exceldata:
        if abbrev not in nren_dict:
            logger.warning(f'{abbrev} unknown. Skipping.')
            continue

        if budget > 200:
            logger.warning(f'{nren} has budget set to >200M EUR for {year}. ({budget})')

        budget_entry = model.BudgetEntry(
            nren=nren_dict[abbrev],
            nren_id=nren_dict[abbrev].id,
            budget=budget,
            year=year
        )
        db.session.merge(budget_entry)
    db.session.commit()


def db_funding_migration(nren_dict):
    # Import the data to database
    data = parse_excel_data.fetch_funding_excel_data()

    for (abbrev, year, client_institution,
            european_funding,
            gov_public_bodies,
            commercial, other) in data:

        _data = [client_institution, european_funding, gov_public_bodies, commercial, other]
        total = sum(_data)
        if not math.isclose(total, 100, abs_tol=0.01) and total != 0:
            logger.warning(f'{abbrev} funding sources for {year} do not sum to 100% ({total})')

        if abbrev not in nren_dict:
            logger.warning(f'{abbrev} unknown. Skipping.')
            continue

        budget_entry = model.FundingSource(
            nren=nren_dict[abbrev],
            nren_id=nren_dict[abbrev].id,
            year=year,
            client_institutions=client_institution,
            european_funding=european_funding,
            gov_public_bodies=gov_public_bodies,
            commercial=commercial,
            other=other)
        db.session.merge(budget_entry)
    db.session.commit()


def db_charging_structure_migration(nren_dict):
    # Import the data to database
    data = parse_excel_data.fetch_charging_structure_excel_data()

    for (abbrev, year, charging_structure) in data:
        if abbrev not in nren_dict:
            logger.warning(f'{abbrev} unknown. Skipping.')
            continue

        charging_structure_entry = model.ChargingStructure(
            nren=nren_dict[abbrev],
            nren_id=nren_dict[abbrev].id,
            year=year,
            fee_type=charging_structure
        )
        db.session.merge(charging_structure_entry)
    db.session.commit()


def db_staffing_migration(nren_dict):
    staff_data = parse_excel_data.fetch_staffing_excel_data()

    nren_staff_map = {}
    for (abbrev, year, permanent_fte, subcontracted_fte) in staff_data:
        if abbrev not in nren_dict:
            logger.warning(f'{abbrev} unknown. Skipping staff data.')
            continue

        nren = nren_dict[abbrev]
        nren_staff_map[(nren.id, year)] = model.NrenStaff(
            nren=nren,
            nren_id=nren.id,
            year=year,
            permanent_fte=permanent_fte,
            subcontracted_fte=subcontracted_fte,
            technical_fte=0,
            non_technical_fte=0
        )

    function_data = parse_excel_data.fetch_staff_function_excel_data()
    for (abbrev, year, technical_fte, non_technical_fte) in function_data:
        if abbrev not in nren_dict:
            logger.warning(f'{abbrev} unknown. Skipping staff function data.')
            continue

        nren = nren_dict[abbrev]
        if (nren.id, year) in nren_staff_map:
            nren_staff_map[(nren.id, year)].technical_fte = technical_fte
            nren_staff_map[(nren.id, year)].non_technical_fte = non_technical_fte
        else:
            nren_staff_map[(nren.id, year)] = model.NrenStaff(
                nren=nren,
                nren_id=nren.id,
                year=year,
                permanent_fte=0,
                subcontracted_fte=0,
                technical_fte=technical_fte,
                non_technical_fte=non_technical_fte
            )

    for nren_staff_model in nren_staff_map.values():
        employed = nren_staff_model.permanent_fte + nren_staff_model.subcontracted_fte
        technical = nren_staff_model.technical_fte + nren_staff_model.non_technical_fte
        if not math.isclose(employed, technical, abs_tol=0.01) and employed != 0 and technical != 0:
            logger.warning(f'{nren_staff_model.nren.name} in {nren_staff_model.year}:'
                           f' FTE do not equal across employed/technical categories ({employed} != {technical})')

        db.session.merge(nren_staff_model)

    db.session.commit()


def db_ecprojects_migration(nren_dict):
    ecproject_data = parse_excel_data.fetch_ecproject_excel_data()
    for (abbrev, year, project) in ecproject_data:
        if abbrev not in nren_dict:
            logger.warning(f'{abbrev} unknown. Skipping.')
            continue

        nren = nren_dict[abbrev]
        ecproject_entry = model.ECProject(nren=nren, nren_id=nren.id, year=year, project=project)
        db.session.merge(ecproject_entry)
    db.session.commit()


def db_organizations_migration(nren_dict):
    organization_data = parse_excel_data.fetch_organization_excel_data()
    for (abbrev, year, org) in organization_data:
        if abbrev not in nren_dict:
            logger.warning(f'{abbrev} unknown. Skipping.')
            continue

        nren = nren_dict[abbrev]
        org_entry = model.ParentOrganization(nren=nren, nren_id=nren.id, year=year, organization=org)
        db.session.merge(org_entry)
    db.session.commit()


def _cli(config, app):
    with app.app_context():
        nren_dict = helpers.get_uppercase_nren_dict()
        db_budget_migration(nren_dict)
        db_funding_migration(nren_dict)
        db_charging_structure_migration(nren_dict)
        db_staffing_migration(nren_dict)
        db_ecprojects_migration(nren_dict)
        db_organizations_migration(nren_dict)


@click.command()
@click.option('--config', type=click.STRING, default='config.json')
def cli(config):
    app_config = load(open(config, 'r'))

    app_config['SQLALCHEMY_BINDS'] = {survey_model.SURVEY_DB_BIND: app_config['SURVEY_DATABASE_URI']}

    app = compendium_v2._create_app_with_db(app_config)
    print("survey-publisher-v1 starting")
    _cli(app_config, app)


if __name__ == "__main__":
    cli()
