import logging
from enum import Enum
from typing import Any, Tuple
from uuid import uuid4

from flask import Blueprint, request
from flask_login import login_required, current_user  # type: ignore
from sqlalchemy import select
from sqlalchemy.orm import lazyload

from compendium_v2.db import db
from compendium_v2.db.model import NREN
from compendium_v2.db.survey_model import Survey, SurveyResponse, SurveyStatus, ResponseStatus, RESPONSE_NOT_STARTED
from compendium_v2.routes import common
from compendium_v2.auth.session_management import admin_required, User


routes = Blueprint('response', __name__)
logger = logging.getLogger(__name__)


SURVEY_RESPONSE_SCHEMA = {
    '$schema': 'http://json-schema.org/draft-07/schema#',
    'type': 'object',
    'properties': {
        'model': {'type': 'object'},
        'locked_by': {'type': 'string'},
        'data': {'type': 'object'},
        'page': {'type': 'number'},
        'verification_status': {'type': 'object'},
        'mode': {'type': 'string'},
        'status': {'type': 'string'},
        'edit_allowed': {'type': 'boolean'},
    },
    'required': ['model', 'locked_by', 'data', 'page', 'verification_status', 'mode', 'status'],
    'additionalProperties': False
}


SURVEY_LOCK_SCHEMA = {
    '$schema': 'http://json-schema.org/draft-07/schema#',
    'type': 'object',
    'properties': {
        'lock_uuid': {'type': 'string'},
        'locked_by': {'type': 'string'},
        'data': {'type': 'object'},
        'verification_status': {'type': 'object'},
        'mode': {'type': 'string'},
        'status': {'type': 'string'},
    },
    'required': ['lock_uuid', 'locked_by', 'data', 'verification_status', 'mode', 'status'],
    'additionalProperties': False
}


SURVEY_STATUS_RESPONSE_SCHEMA = {
    '$schema': 'http://json-schema.org/draft-07/schema#',
    'type': 'object',
    'properties': {
        'locked_by': {'type': 'string'},
        'mode': {'type': 'string'},
        'status': {'type': 'string'},
    },
    'required': ['locked_by', 'mode', 'status'],
    'additionalProperties': False
}


class VerificationStatus(str, Enum):
    Unverified = "unverified"     # a question that has its answered copied from last year
    Verified = "verified"         # a question for which last years answer was verified
    Edited = "edited"             # a question for which last years answer was edited


class SurveyMode(str, Enum):
    Display = "display"
    Edit = "edit"


def check_access_nren(user: User, nren: str) -> bool:
    if user.is_anonymous:
        return False
    if user.is_admin:
        return True
    if nren == user.nren:
        return True
    return False


@routes.route('/try/<int:year>', methods=['GET'])
@common.require_accepts_json
@admin_required
def try_survey(year) -> Any:
    """
    Get a survey without any associated nren for trying out the survey.
    The survey will behave exactly as when an NREN opens it.

    response will either a 400 range code with a descriptive message or a 200 code with the following format:

    .. asjson::
        compendium_v2.routes.response.SURVEY_RESPONSE_SCHEMA

    """
    survey = db.session.scalar(select(Survey).where(Survey.year == year))
    if not survey:
        return {'message': 'Survey not found'}, 404

    return {
        "model": survey.survey,
        "locked_by": "",
        "data": {},
        "page": 0,
        "verification_status": {},
        "mode": SurveyMode.Edit,
        "status": RESPONSE_NOT_STARTED,
        "edit_allowed": True
    }


@routes.route('/inspect/<int:year>', methods=['GET'])
@common.require_accepts_json
@admin_required
def inspect_survey(year) -> Any:
    """
    Get a survey without any associated nren for inspecting all questions.
    All questions are made visible and any VisibleIf condition added to the title.

    response will either a 400 range code with a descriptive message or a 200 code with the following format:

    .. asjson::
        compendium_v2.routes.response.SURVEY_RESPONSE_SCHEMA

    """
    survey = db.session.scalar(select(Survey).where(Survey.year == year))
    if not survey:
        return {'message': 'Survey not found'}, 404

    def visible_visitor(object, items):
        for key, value in items:
            if isinstance(value, dict):
                visible_visitor(value, value.items())
            elif isinstance(value, list):
                visible_visitor(value, enumerate(value))
            elif key == 'visibleIf':
                object['title'] = object['title'] + ' (visibleif: [' + value.replace('{', '#').replace('}', '#') + '])'
                object[key] = 'true'

    visible_visitor(survey.survey, survey.survey.items())

    return {
        "model": survey.survey,
        "locked_by": "",
        "data": {},
        "page": 0,
        "verification_status": {},
        "mode": SurveyMode.Edit,
        "status": RESPONSE_NOT_STARTED,
        "edit_allowed": True
    }


def get_response_data(response, year, nren_id) -> Tuple[dict, int, dict, str]:
    data = {}
    page = 0
    verification_status = {}
    locked_by = response.lock_username if response else ""

    if response and response.answers:
        data = response.answers["data"]
        page = response.answers["page"]
        verification_status = response.answers["verification_status"]
    else:
        previous_response = db.session.scalar(
            select(SurveyResponse).where(SurveyResponse.survey_year == year-1).where(SurveyResponse.nren_id == nren_id)
        )
        if previous_response:
            # TODO add a 'migration' hook here for updating data per year
            previous_response_data: dict = previous_response.answers["data"]
            data = previous_response_data
            verification_status = {question_name: VerificationStatus.Unverified for question_name in data.keys()}

    return data, page, verification_status, locked_by


@routes.route('/load/<int:year>/<string:nren_name>', methods=['GET'])
@common.require_accepts_json
@login_required
def load_survey(year, nren_name) -> Any:
    """
    Get a survey for an nren.
    If the survey was saved before, that data will be loaded.
    If the survey was not saved before and the survey was completed last year,
    the data from last year will be prefilled.

    response will either a 400 range code with a descriptive message or a 200 code with the following format:

    .. asjson::
        compendium_v2.routes.response.SURVEY_RESPONSE_SCHEMA

    """
    nren = db.session.scalar(select(NREN).filter(NREN.name == nren_name))
    if not nren:
        return {'message': 'NREN not found'}, 404

    survey = db.session.scalar(select(Survey).where(Survey.year == year))
    if not survey:
        return {'message': 'Survey not found'}, 404

    if not check_access_nren(current_user, nren):
        return {'message': 'You do not have permissions to access this survey.'}, 403

    response = db.session.scalar(
        select(SurveyResponse).where(SurveyResponse.survey_year == year).where(SurveyResponse.nren_id == nren.id)
    )

    data, page, verification_status, locked_by = get_response_data(response, year, nren.id)

    return {
        "model": survey.survey,
        "locked_by": locked_by,
        "data": data,
        "page": page,
        "verification_status": verification_status,
        "mode": SurveyMode.Display,
        "status": response.status.value if response else RESPONSE_NOT_STARTED,
        "edit_allowed": current_user.is_admin or survey.status == SurveyStatus.open
    }


@routes.route('/lock/<int:year>/<string:nren_name>', methods=['POST'])
@common.require_accepts_json
@login_required
def lock_survey(year, nren_name) -> Any:
    """
    Get the edit lock on a response. Creates the response if it doesnt exist yet.
    It also includes the survey data to prevent race conditions.

    response will either a 400 range code with a descriptive message or a 200 code with the following format:

    .. asjson::
        compendium_v2.routes.response.SURVEY_LOCK_SCHEMA

    """
    nren = db.session.scalar(select(NREN).filter(NREN.name == nren_name))
    if not nren:
        return {'message': 'NREN not found'}, 404

    survey = db.session.scalar(select(Survey).where(Survey.year == year))
    if not survey:
        return {'message': 'Survey not found'}, 404

    if not check_access_nren(current_user, nren):
        return {'message': 'You do not have permissions to access this survey.'}, 403

    if survey.status != SurveyStatus.open and not current_user.is_admin:
        return {'message': 'Survey is closed'}, 400

    response = db.session.scalar(
        select(SurveyResponse).where(SurveyResponse.survey_year == year)
                              .where(SurveyResponse.nren_id == nren.id)
                              .options(lazyload("*"))
                              .with_for_update()
    )

    data, page, verification_status, locked_by = get_response_data(response, year, nren.id)

    # it's possible this is the first time somebody opens the survey for edit, in
    # that case we have to create the response:
    if response is None:
        response = SurveyResponse(survey_year=year, nren_id=nren.id, status=ResponseStatus.started, answers={})
        db.session.add(response)

    if response.locked_by is not None:
        return {'message': 'This survey is already locked.'}, 403

    lock_uuid = uuid4()
    response.lock_uuid = lock_uuid
    response.locked_by = current_user.id

    db.session.commit()

    return {
        'lock_uuid': lock_uuid,
        'locked_by': response.lock_username,
        'data': data,
        'verification_status': verification_status,
        'mode': SurveyMode.Edit,
        'status': response.status.value
    }


@routes.route('/save/<int:year>/<string:nren_name>', methods=['POST'])
@common.require_accepts_json
@login_required
def save_survey(year, nren_name) -> Any:
    """
    endpoint to save a survey response

    expects the following entries in the json body:
    "lock_uuid" to verify if the user has the edit lock;
    "new_state" to distinguish the save/save and stop editing/complete actions;
    "data", "page", "verification_status" which are stored in the db.

    response will either a 400 range code with a descriptive message or a 200 code with the following format:

    .. asjson::
        compendium_v2.routes.response.SURVEY_STATUS_RESPONSE_SCHEMA

    """

    class NewState(str, Enum):
        ReadOnly = "readonly"
        Editing = "editing"
        Completed = "completed"

    nren = db.session.scalar(select(NREN).filter(NREN.name == nren_name))
    if nren is None:
        return {'message': 'NREN not found'}, 404

    survey = db.session.scalar(select(Survey).where(Survey.year == year))
    if survey is None:
        return {'message': 'Survey not found'}, 404

    if not check_access_nren(current_user, nren):
        return {'message': 'You do not have permission to edit this survey.'}, 403

    if survey.status != SurveyStatus.open and not current_user.is_admin:
        return {'message': 'Survey is closed'}, 400

    response = db.session.scalar(
        select(SurveyResponse).where(SurveyResponse.survey_year == year).where(SurveyResponse.nren_id == nren.id)
    )
    if response is None:
        return {'message': 'Survey response not found'}, 400

    save_survey = request.json
    if not save_survey:
        return {'message': 'Invalid Survey Format'}, 400

    lock_uuid = save_survey["lock_uuid"]
    if (lock_uuid != str(response.lock_uuid)):
        return {'message': 'Somebody else is currently editing this survey.'}, 403

    new_state = save_survey["new_state"]
    if (new_state == NewState.Editing):
        pass
    elif (new_state == NewState.Completed):
        response.locked_by = None
        response.status = ResponseStatus.completed
    elif (new_state == NewState.ReadOnly):
        response.locked_by = None
    else:
        return {'message': 'Invalid new_state paramater'}, 400

    response.answers = {
        "data": save_survey["data"],
        "page": save_survey["page"],
        "verification_status": save_survey["verification_status"]
    }
    db.session.commit()

    mode = SurveyMode.Edit if new_state == NewState.Editing else SurveyMode.Display
    return {'locked_by': response.lock_username, 'mode': mode, 'status': response.status.value}


@routes.route('/unlock/<int:year>/<string:nren_name>', methods=['POST'])
@common.require_accepts_json
@login_required
def unlock_survey(year, nren_name) -> Any:
    """
    endpoint to release a lock on a survey response without saving any data

    response will either a 400 range code with a descriptive message or a 200 code with the following format:

    .. asjson::
        compendium_v2.routes.response.SURVEY_STATUS_RESPONSE_SCHEMA

    """
    nren = db.session.scalar(select(NREN).filter(NREN.name == nren_name))
    if nren is None:
        return {'message': 'NREN not found'}, 404

    survey = db.session.scalar(select(Survey).where(Survey.year == year))
    if survey is None:
        return {'message': 'Survey not found'}, 404

    if not check_access_nren(current_user, nren):
        return {'message': 'You do not have permission to edit this survey.'}, 403

    response = db.session.scalar(
        select(SurveyResponse).where(SurveyResponse.survey_year == year).where(SurveyResponse.nren_id == nren.id)
    )
    if response is None:
        return {'message': 'Survey response not found'}, 404

    if current_user.id != response.locked_by and not current_user.is_admin:
        return {'message': 'This survey was not locked by you.'}, 403

    response.locked_by = None  # we leave lock_uuid, it will be overwritten when a new lock is acquired
    db.session.commit()

    return {'locked_by': response.lock_username, 'mode': SurveyMode.Display, 'status': response.status.value}
