import logging
from typing import Any, TypedDict, List, Dict

from flask import Blueprint
from sqlalchemy import select
from sqlalchemy.orm import joinedload, load_only

from compendium_v2.db import db
from compendium_v2.db.model import NREN
from compendium_v2.db.survey_model import Survey, SurveyResponse, SurveyStatus, ResponseStatus, RESPONSE_NOT_STARTED
from compendium_v2.routes import common
from compendium_v2.auth.session_management import admin_required


routes = Blueprint('survey', __name__)
logger = logging.getLogger(__name__)


LIST_SURVEYS_RESPONSE_SCHEMA = {
    '$schema': 'http://json-schema.org/draft-07/schema#',
    'definitions': {
        'response': {
            'type': 'object',
            'properties': {
                'nren': {'type': 'string'},
                'status': {'type': 'string'},
                'lock_description': {'type': 'string'},
            },
            'required': ['nren', 'status', 'lock_description'],
            'additionalProperties': False
        },
        'survey': {
            'type': 'object',
            'properties': {
                'year': {'type': 'integer'},
                'status': {'type': 'string'},
                'responses': {'type': 'array', 'items': {'$ref': '#/definitions/response'}},
            },
            'required': ['year', 'status', 'responses'],
            'additionalProperties': False
        }
    },
    'type': 'array',
    'items': {'$ref': '#/definitions/survey'}
}


@routes.route('/list', methods=['GET'])
@common.require_accepts_json
@admin_required
def list_surveys() -> Any:
    """
    retrieve a list of surveys and responses, including their status

    response will be formatted as:

    .. asjson::
        compendium_v2.routes.survey.LIST_SURVEYS_RESPONSE_SCHEMA

    """
    surveys = db.session.scalars(
        select(Survey).options(
            load_only(Survey.year, Survey.status),
            joinedload(Survey.responses).load_only(SurveyResponse.status)
            .joinedload(SurveyResponse.nren).load_only(NREN.name)
        ).order_by(Survey.year.desc())
    ).unique()

    def response_key(response):
        return response.status.value + response.nren.name.lower()

    class SurveyDict(TypedDict):
        year: int
        status: str
        responses: List[Dict[str, str]]

    entries: List[SurveyDict] = [
        {
            "year": entry.year,
            "status": entry.status.value,
            "responses": [
                {
                    "nren": r.nren.name,
                    "status": r.status.value,
                    "lock_description": r.lock_description
                }
                for r in sorted(entry.responses, key=response_key)
            ]
        }
        for entry in surveys
    ]

    # add in nrens without a response if the survey is open
    nren_names = set([name for name in db.session.scalars(select(NREN.name))])
    for entry in entries:
        if entry["status"] != SurveyStatus.open.value:
            continue
        nrens_with_responses = set([r["nren"] for r in entry["responses"]])
        for nren_name in sorted(nren_names.difference(nrens_with_responses), key=str.lower):
            entry["responses"].append({"nren": nren_name, "status": RESPONSE_NOT_STARTED, "lock_description": ""})

    return entries


@routes.route('/new', methods=['POST'])
@common.require_accepts_json
@admin_required
def start_new_survey() -> Any:
    """
    endpoint to initiate a new survey

    :returns: ``{'success': True}`` or a 400 or 404 status with a descriptive message
    """
    all_surveys = db.session.scalars(select(Survey).options(load_only(Survey.status)))
    if any([survey.status != SurveyStatus.published for survey in all_surveys]):
        return {'message': 'All earlier surveys should be published before starting a new one'}, 400

    last_survey = db.session.scalar(
        select(Survey).order_by(Survey.year.desc()).limit(1)
    )

    if not last_survey:
        return {'message': 'No surveys found'}, 404

    new_year = last_survey.year + 1
    new_survey = last_survey.survey
    new_survey = Survey(year=new_year, survey=new_survey, status=SurveyStatus.closed)
    db.session.add(new_survey)
    db.session.commit()

    return {'success': True}


@routes.route('/open/<int:year>', methods=['POST'])
@common.require_accepts_json
@admin_required
def open_survey(year) -> Any:
    """
    endpoint to open a survey to the nrens

    :returns: ``{'success': True}`` or a 400 or 404 status with a descriptive message
    """
    survey = db.session.scalar(select(Survey).where(Survey.year == year))
    if not survey:
        return {'message': 'Survey not found'}, 404

    if survey.status != SurveyStatus.closed:
        return {'message': 'Survey is not closed and can therefore not be opened'}, 400

    all_surveys = db.session.scalars(select(Survey))
    if any([s.status == SurveyStatus.open for s in all_surveys]):
        return {'message': 'There already is an open survey'}, 400

    survey.status = SurveyStatus.open
    db.session.commit()

    return {'success': True}


@routes.route('/close/<int:year>', methods=['POST'])
@common.require_accepts_json
@admin_required
def close_survey(year) -> Any:
    """
    endpoint to close a survey to the nrens

    :returns: ``{'success': True}`` or a 400 or 404 status with a descriptive message
    """
    survey = db.session.scalar(select(Survey).where(Survey.year == year))
    if not survey:
        return {'message': 'Survey not found'}, 404

    if survey.status != SurveyStatus.open:
        return {'message': 'Survey is not open and can therefore not be closed'}, 400

    survey.status = SurveyStatus.closed
    db.session.commit()

    return {'success': True}


@routes.route('/publish/<int:year>', methods=['POST'])
@common.require_accepts_json
@admin_required
def publish_survey(year) -> Any:
    """
    endpoint to publish a survey to the compendium website

    :returns: ``{'success': True}`` or a 400 or 404 status with a descriptive message
    """
    survey = db.session.scalar(select(Survey).where(Survey.year == year))
    if not survey:
        return {'message': 'Survey not found'}, 404

    if survey.status not in [SurveyStatus.closed, SurveyStatus.published]:
        return {'message': 'Survey is not closed or published and can therefore not be published'}, 400

    # TODO probably replace checked state with something else
    if any([response.status != ResponseStatus.checked for response in survey.responses]):
        return {'message': 'There are responses that arent checked yet'}, 400

    # TODO call new survey_publisher with all responses and the year

    survey.status = SurveyStatus.published
    db.session.commit()

    return {'success': True}
