import React, { useState, useEffect } from "react";
import Accordion from 'react-bootstrap/Accordion';
import Button from 'react-bootstrap/Button';
import Table from 'react-bootstrap/Table';
import { useNavigate } from 'react-router-dom';
import { ResponseStatus, SurveyStatus } from "./Schema";


async function fetchSurveys(): Promise<Survey[]> {
    try {
        const response = await fetch('/api/survey/list');
        const userList = await response.json();
        return userList
    } catch (error) {
        console.log('failed fetching survey list..');
        return [];
    }
}

interface Response {
    nren: string
    status: string
    lock_description: string
}

interface Survey {
    year: number
    status: string
    responses: Response[]
}

function SurveyManagementComponent() {
    const [surveys, setSurveys] = useState<Survey[]>([]);

    useEffect(() => {
        fetchSurveys().then((surveyList) => {
            setSurveys(surveyList);
        });
    }, []);

    async function newSurvey() {
        await fetch('/api/survey/new', { method: 'POST' });
        fetchSurveys().then((surveyList) => {
            setSurveys(surveyList);
        });
    }

    async function postSurveyStatus(year, status) {
        await fetch('/api/survey/' + status + '/' + year, { method: 'POST' });
        fetchSurveys().then((surveyList) => {
            setSurveys(surveyList);
        });
    }

    async function removeLock(year, nrenName) {
        await fetch('/api/response/unlock/' + year + '/' + nrenName, { method: 'POST' });
        fetchSurveys().then((surveyList) => {
            setSurveys(surveyList);
        });
    }

    const newSurveyAllowed = surveys.length > 0 && surveys.every(s => s.status == SurveyStatus.published);

    const navigate = useNavigate();

    return (
        <div>
            <Button onClick={newSurvey} disabled={!newSurveyAllowed} style={{ pointerEvents: 'auto' }}
                title="Create a new survey for the next year. Only possible if all current surveys are published.">
                start new survey
            </Button>
            <Accordion>
                {surveys.map(survey => (
                    <Accordion.Item eventKey={survey.year.toString()} key={survey.year}>
                        <Accordion.Header>{survey.year} - {survey.status}</Accordion.Header>
                        <Accordion.Body>
                            <div style={{ marginLeft: '.5rem' }}>
                                <Button style={{ marginLeft: '.5rem' }} onClick={() => navigate(`/survey/admin/inspect/${survey.year}`)}
                                    title="Open the survey for inspection with all questions visible and any visibleIf logic added to the title.">
                                    Inspect Survey
                                </Button>
                                <Button style={{ marginLeft: '.5rem' }} onClick={() => navigate(`/survey/admin/try/${survey.year}`)}
                                    title="Open the survey exactly as the nrens will see it, but without any nren data.">
                                    Try Survey
                                </Button>
                                {survey.status !== 'published' &&
                                    <Button onClick={() => postSurveyStatus(survey.year, 'open')} disabled={survey.status != SurveyStatus.closed} style={{ pointerEvents: 'auto', marginLeft: '.5rem' }}
                                        title="Allow the NRENs to respond to this survey. Only 1 survey may be open at a time, and published surveys cannot be opened anymore.">
                                        Mark as Open
                                    </Button>}
                                <Button onClick={() => postSurveyStatus(survey.year, 'close')} disabled={survey.status != SurveyStatus.open} style={{ pointerEvents: 'auto', marginLeft: '.5rem' }}
                                    title="Do not allow the NRENs to respond to this survey anymore. Only surveys with status open can be closed.">
                                    Mark as closed
                                </Button>
                                <Button onClick={() => postSurveyStatus(survey.year, 'publish')} disabled={(survey.status != SurveyStatus.closed && survey.status != SurveyStatus.published) || !survey.responses.every(r => r.status == ResponseStatus.checked)} style={{ pointerEvents: 'auto', marginLeft: '.5rem' }}
                                    title="Publish or re-publish all survey responses to the compendium website. This is only possible if the survey is closed or published already, and all responses are checked.">
                                    Publish results
                                </Button>
                                {/* maybe something like this: <Button disabled={survey.status!='published'}>Show in compendium</Button>
                        <Button disabled={survey.status!='published'}>Hide from compendium</Button> */}
                            </div>
                            <Table>
                                <tbody>
                                    {survey.responses.map(response => (
                                        <tr key={response.nren}>
                                            <td>{response.nren}</td>
                                            <td>{response.status}</td>
                                            <td>{response.lock_description}</td>
                                            <td>
                                                <Button onClick={() => navigate(`/survey/response/${survey.year}/${response.nren}`)} style={{ pointerEvents: 'auto' }}
                                                    title="Open the responses of the NREN.">
                                                    open
                                                </Button>
                                                <Button onClick={() => removeLock(survey.year, response.nren)} disabled={response.lock_description == ''} style={{ pointerEvents: 'auto' }}
                                                    title="Remove the lock from the survey so that another person can open the survey for editing. WARNING: The person that currently has the lock will not be abe to save their changes anymore once someone else starts editing!">
                                                    remove lock
                                                </Button>
                                            </td>
                                        </tr>
                                    ))}

                                </tbody>
                            </Table>

                        </Accordion.Body>
                    </Accordion.Item>
                ))}
            </Accordion>
        </div>
    );

}

export default SurveyManagementComponent;
