import React, { useState, useEffect, useContext } from "react";
import { Container, Row, Table } from "react-bootstrap";
import { userContext } from "./providers/UserProvider";
import { User, Nren } from "./Schema";


async function fetchUsers(): Promise<User[]> {
    try {
        const response = await fetch('/api/user/list');
        const userList = await response.json();
        return userList
    } catch (error) {
        console.log('handle this better..');
        return [];
    }
}

async function fetchNrens(): Promise<Nren[]> {
    try {
        const response = await fetch('/api/nren/list');
        const userList = await response.json();
        return userList
    } catch (error) {
        console.log('handle this better..');
        return [];
    }
}

const updateUser = async (id, changes) => {

    const body = {
        id: id,
        ...changes
    }
    const requestOptions = {
        method: 'PUT',
        headers: {
            'Content-Type': 'application/json',
        },
        body: JSON.stringify(body),
    };

    const response = await fetch(`/api/user/${id}`, requestOptions)
    const data = await response.json()
    if (!response.ok) {
        throw new Error(data.message);
    }

    return data.user;
};


function UserManagementComponent() {
    const [users, setUsers] = useState<User[]>([]);
    const [nrens, setNrens] = useState<Nren[]>([]);
    const { user: loggedInUser } = useContext(userContext);

    useEffect(() => {
        fetchUsers().then((userList) => {
            setUsers(userList);
        });

        fetchNrens().then((nrenList) => {
            setNrens(nrenList.sort((a, b) => a.name.localeCompare(b.name)))
        })
    }, []);

    const handleInputChange = (event: React.ChangeEvent<HTMLInputElement | HTMLSelectElement>, user: User) => {
        const index = users.findIndex((u) => u.id === user.id);
        const updatedUsers = [...users];

        const { name } = event.target;

        const update = {};

        if (name === 'active') {
            update[name] = (event.target as HTMLInputElement).checked
        } else {
            update[name] = event.target.value;
        }

        updateUser(user.id, update).then((user) => {
            updatedUsers[index] = user;
            setUsers(updatedUsers);
        }).catch((error) => {
            alert(error.message);
        });
    };

    const findNren = (value) => {
        return nrens.find((nren) => {
            return nren.id == value || nren.name == value
        })?.id
    }

    return (
        <Container style={{ maxWidth: '90vw', }}>
            <Row>
                <h1> User Management Page</h1>
                <Table>
                    <thead>
                        <tr>
                            <th className='pt-3' style={{ border: "1px solid #ddd" }}>Id</th>
                            <th className='pt-3' style={{ border: "1px solid #ddd" }}>Active</th>
                            <th className='pt-3' style={{ border: "1px solid #ddd" }}>Role</th>
                            <th className='pt-3' style={{ border: "1px solid #ddd" }}>Email</th>
                            <th className='pt-3' style={{ border: "1px solid #ddd" }}>Full Name</th>
                            <th className='pt-3' style={{ border: "1px solid #ddd" }}>OIDC Sub</th>
                            <th className='pt-3' style={{ border: "1px solid #ddd" }}>NREN</th>
                        </tr>
                    </thead>
                    <tbody>
                        {users.map(user => (
                            <tr key={user.id}>
                                <td style={{ border: "1px dotted #ddd" }}>{user.id}</td>
                                <td style={{ border: "1px dotted #ddd" }}>
                                    {user.id == loggedInUser.id ? 'Active' : <input
                                        type="checkbox"
                                        name="active"
                                        checked={user.permissions.active}
                                        onChange={(event) => handleInputChange(event, user)}
                                    />}

                                </td>
                                <td style={{ border: "1px dotted #ddd" }}>
                                    {user.id == loggedInUser.id ? (user.role.charAt(0).toUpperCase()
                                        + user.role.slice(1)) : <select
                                            name="role"
                                            defaultValue={user.role}
                                            onChange={(event) => handleInputChange(event, user)}>
                                        <option value="admin">Admin</option>
                                        <option value="user">User</option>
                                    </select>}
                                </td>

                                <td style={{ border: "1px dotted #ddd" }}>
                                    {user.email}
                                </td>
                                <td style={{ border: "1px dotted #ddd" }}>
                                    {user.name}
                                </td>
                                <td style={{ border: "1px dotted #ddd" }}>
                                    {user.oidc_sub}
                                </td>
                                <td style={{ border: '1px dotted #ddd' }}>
                                    <select
                                        name="nren"
                                        multiple={false}
                                        defaultValue={user.nrens.length > 0 ? findNren(user.nrens[0]) : undefined}
                                        onChange={(event) => handleInputChange(event, user)}>
                                        <option>
                                            Select NREN
                                        </option>
                                        {nrens.map((nren) => (
                                            <option key={nren.id} value={nren.id}>
                                                {nren.name}
                                            </option>
                                        ))}
                                    </select>
                                </td>
                            </tr>
                        ))}
                    </tbody>

                </Table>
            </Row>

        </Container>
    );

}

export default UserManagementComponent;
